from pypicloud_client.client import PypicloudClient

from loktar.cmd import exe
from loktar.cmd import exec_command_with_retry
from loktar.environment import MAX_RETRY_PYTHON_UPLOAD
from loktar.exceptions import WheelPackageFail
from loktar.plugin import ComplexPlugin


def run(*args, **kwargs):
    Whl(args[0]).run()


class Whl(ComplexPlugin):
        def __init__(self, package_info):
            ComplexPlugin.__init__(self, package_info,
                                   {
                                       "command": {
                                           "run": "",
                                           "clean": "make clean"
                                       }
                                   })
            self.timeline = {
                10: self.get_version,
                40: self.release
            }
            self.pypicloud = PypicloudClient()

        def run(self):
            """Default method for running the timeline
            """
            self._run()

        def get_version(self):
            pypicloud = PypicloudClient()
            if self.package_info["mode"] == "master":
                self.share_memory["latest_version"] = str(int(filter(
                    lambda pkg: pkg.isdigit(),
                    sorted(
                        map(
                            lambda pkg: pkg["version"],
                            pypicloud.get_versions(self.package_info["pkg_name"])
                        ),
                        reverse=True
                    )
                )[0]) + 1)
            else:
                self.share_memory["latest_version"] = self.package_info["mode"]

        def release(self):
            """
            """
            with self.cwd(self.path):
                # Edit the package version
                if not exe("sed -i 's/VERSION = .*/VERSION = \"{0}\"/g' setup.py"
                           .format(self.share_memory["latest_version"]), remote=False):
                    raise WheelPackageFail("Failed to update the version")

                # Build package
                if not exec_command_with_retry("make release", max_retry=MAX_RETRY_PYTHON_UPLOAD, remote=False):
                    raise WheelPackageFail('Release failed')

                self.logger.info('Wheel package built & uploaded')



from fabric.api import local
from fabric.api import run
from fabric.api import settings

from loktar.log import Log


def exec_command_with_retry(cmd, remote, max_retry):
    """Execute and retry a command"""
    with settings(warn_only=True):
        id_try = 0
        launch = local if remote == 0 else run
        while id_try < max_retry:
            result = launch(cmd)
            if result.succeeded:
                return True
            else:
                id_try += 1
        logger = Log()
        logger.error("The command : {0} failed after {1} retries".format(cmd, max_retry))
        return False


def exe(cmd, remote=True):
    """Execute a command

    Args:
        cmd (str): Command to execute
        remote (bool): Give the context execution remote (True) or local (False)

    Returns:
        bool: True if everything went well, False otherwise
    """
    launch = run if remote else local
    with settings(warn_only=True):
        result = launch(cmd)
        if result.failed:
            logger = Log()
            logger.error(result)
            return False
        return True



from fabric.api import env
from fabric.api import lcd
from fabric.api import local
from fabric.exceptions import NetworkError
import json
import os
from uuid import uuid4

from loktar.cmd import exec_command_with_retry
from loktar.exceptions import PrepareEnvFail
from loktar.log import Log

env.disable_known_hosts = True
env.warn_only = True

PREFIX_ENV_VAR = "LOKTAR_"

# DEBIAN PACKAGE INFORMATION
CODE_NAME = os.getenv("{0}CODE_NAME".format(PREFIX_ENV_VAR), "rc")
COMPENANT = os.getenv("{0}COMPENANT".format(PREFIX_ENV_VAR), "main")
DEBIAN_REPOSITORY = os.getenv("{0}DEBIAN_REPOSITORY".format(PREFIX_ENV_VAR), None)

# About retry on command or http requests ...
MAX_RETRY_PYTHON_UPLOAD = os.getenv("{0}MAX_RETRY_PYTHON_UPLOAD".format(PREFIX_ENV_VAR), 3)
MAX_RETRY_DEBIAN_UPLOAD = os.getenv("{0}MAX_RETRY_DEBIAN_UPLOAD".format(PREFIX_ENV_VAR), 3)
MAX_RETRY_DOCKER_BUILD = os.getenv("{0}MAX_RETRY_DOCKER_BUILD".format(PREFIX_ENV_VAR), 3)
MAX_RETRY_DOCKER_PUSH = os.getenv("{0}MAX_RETRY_DOCKER_PUSH".format(PREFIX_ENV_VAR), 3)
MAX_RETRY_GITHUB = os.getenv("{0}MAX_RETRY_GITHUB".format(PREFIX_ENV_VAR), 3)

CONSUL = os.getenv("CONSUL", None)

GITHUB_INFO = {
    "login": {
        "user": os.getenv("{0}GITHUB_INFO_LOGIN_USER".format(PREFIX_ENV_VAR), None),
        "password": os.getenv("{0}GITHUB_INFO_LOGIN_PASSWORD".format(PREFIX_ENV_VAR), None)
    },
    "organization": os.getenv("{0}GITHUB_INFO_ORGANIZATION".format(PREFIX_ENV_VAR), None),
    "repository": os.getenv("{0}GITHUB_INFO_REPOSITORY".format(PREFIX_ENV_VAR), None),
    "notification": {
        "pending": os.getenv("{0}GITHUB_INFO_NOTIFICATION_PENDING".format(PREFIX_ENV_VAR),
                             "The build is running. If i was you I would leave quickly"),
        "success": os.getenv("{0}GITHUB_INFO_NOTIFICATION_SUCCESS".format(PREFIX_ENV_VAR),
                             "The build succeded! You're not so bad after all"),
        "error": os.getenv("{0}GITHUB_INFO_NOTIFICATION_ERROR".format(PREFIX_ENV_VAR),
                           "Error! I won't lie, you're really bad"),
        "failure": os.getenv("{0}GITHUB_INFO_NOTIFICATION_FAILLURE".format(PREFIX_ENV_VAR),
                             "Failure. Nobody is happy."),
        "unknown": os.getenv("{0}GITHUB_INFO_NOTIFICATION_UNKNOWN".format(PREFIX_ENV_VAR),
                             "Unknown state, so 42.")
    },
    "default_master_branch": os.getenv("{0}GITHUB_DEFAULT_MASTER_BRANCH".format(PREFIX_ENV_VAR), "master"),
    "manifest_path": os.getenv("{0}GITHUB_MANIFEST_PATH".format(PREFIX_ENV_VAR), "Manifest")
}
GITHUB_TOKEN = os.getenv("{0}GITHUB_TOKEN".format(PREFIX_ENV_VAR), None)
GITHUB_API_COMMITS = {
    "url": "https://api.github.com/repos/{0}/{1}/git/commits".format(GITHUB_INFO["organization"],
                                                                     GITHUB_INFO["repository"]),
    "headers": {
        "Authorization": "token {0}",
        "Accept": os.getenv("{0}GITHUB_API_COMMITS_HEADERS_ACCEPT".format(PREFIX_ENV_VAR),
                            "application/vnd.github-commitcomment.full+json")
    }
}
CONFIG_CI_FILE = os.getenv("{0}CONFIG_CI_FILE".format(PREFIX_ENV_VAR), "config.json")
DOCKER_ENGINE = {
    "address": os.getenv("{0}DOCKER_CI_ADDRESS", "127.0.0.1"),
    "port": os.getenv("{0}DOCKER_CI_PORT", 2375),
    "registry_url": os.getenv("{0}DOCKER_CI_REGISTRY", None)
}
DOCKER_CI = {
    "registry": os.getenv("{0}DOCKER_CI_REGISTRY".format(PREFIX_ENV_VAR), None),
    "image": os.getenv("{0}DOCKER_CI_IMAGE".format(PREFIX_ENV_VAR), None),
    "acceptance_image": os.getenv("{0}DOCKER_CI_IMAGE".format(PREFIX_ENV_VAR), None),
    "base_image": os.getenv("{0}DOCKER_CI_BASE_IMAGE".format(PREFIX_ENV_VAR), None),
    "user": os.getenv("{0}DOCKER_CI_USER".format(PREFIX_ENV_VAR), None),
    "address_build": os.getenv("{0}DOCKER_CI_ADDRESS_BUILD".format(PREFIX_ENV_VAR), None),
    "services": os.getenv("{0}DOCKER_CI_SERVICES".format(PREFIX_ENV_VAR), None),
}
ROOT_PATH = {
    "jenkins": os.getenv("{0}ROOT_PATH_JENKINS".format(PREFIX_ENV_VAR), "/var/lib/jenkins"),
    "container": os.getenv("{0}ROOT_PATH_CONTAINER".format(PREFIX_ENV_VAR), None),
    "build": os.getenv("{0}ROOT_PATH_BUILD".format(PREFIX_ENV_VAR), None)
}
BUILD_PACKAGE = {
    "wheel": os.getenv("{0}BUILD_PACKAGE_WHEEL".format(PREFIX_ENV_VAR), "python setup.py bdist_wheel"),
    "debian": os.getenv("{0}BUILD_PACKAGE_DEBIAN".format(PREFIX_ENV_VAR), "dpkg --build"),
    "docker": os.getenv("{0}BUILD_PACKAGE_DOCKER".format(PREFIX_ENV_VAR),
                        "docker build --no-cache -t {0}".format(DOCKER_ENGINE["registry_url"]))
}

PUSH_PACKAGE = {
    "pypicloud": os.getenv("{0}PUSH_PACKAGE_PYPICLOUD".format(PREFIX_ENV_VAR), "twine upload -r pypicloud *whl"),
    "debian": os.getenv("{0}PUSH_PACKAGE_DEBIAN".format(PREFIX_ENV_VAR),
                        "deb-s3 upload -p -b {0} -c {1} -m {2} -v public -a amd64 *.deb"
                        .format(DEBIAN_REPOSITORY, CODE_NAME, COMPENANT)),
    "docker": os.getenv("{0}PUSH_PACKAGE_DOCKER".format(PREFIX_ENV_VAR),
                        "docker push {0}".format(DOCKER_ENGINE["registry_url"]))
}

JENKINS = {
    "host": "delivery.net:8080",
    "user": "lokbot",
    "password": "L39yN@;y6+We"
}

SLAVE_ENVIRONMENT = {
    "PATH": "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/usr/games:/usr/local/games:{0}/.local/bin"
            .format(ROOT_PATH["container"]),
    "env_file": os.getenv("{0}ENV_FILE".format(PREFIX_ENV_VAR), None)
}

AWS = {
    "REGION": os.getenv("{0}AWS_REGION".format(PREFIX_ENV_VAR), None),
    "BUCKET": os.getenv("{0}AWS_BUCKET".format(PREFIX_ENV_VAR), None)
}

PLUGINS_LOCATIONS = os.getenv("{0}PLUGINS_LOCATIONS".format(PREFIX_ENV_VAR), [])


def prepare_test_env(branch, github_organization=None, github_repository=None):
    """Prepare the test environment

    Args:
        branch (str): Name of the branch the repository should be checkout to.
        github_organization (str): this is the github organization for get back the repository, default value None.
                                   Also can be set by environment variable LOKTAR_GITHUB_INFO_ORGANIZATION
        github_repository (str): this is the target repository to download, default value None
                                 Also can be set by environment variable LOKTAR_GITHUB_INFO_REPOSITORY

    Raises:
        PrepareEnvFail: Failed to prepare the environment.
    """
    logger = Log()
    unique_name_dir = str(uuid4())
    unique_path_dir = "/tmp/{0}".format(unique_name_dir)

    logger.info("Preparing the test environment")

    os.mkdir(unique_path_dir)
    try:
        if not exec_command_with_retry("git clone -b {0} --single-branch git@github.com:{1}/{2}.git {3}"
                                       .format(branch,
                                               GITHUB_INFO["organization"]
                                               if github_organization is None else
                                               github_organization,
                                               GITHUB_INFO["repository"]
                                               if github_repository is None else
                                               github_repository,
                                               unique_path_dir),
                                       0,
                                       MAX_RETRY_GITHUB):
            raise PrepareEnvFail

        with lcd(unique_path_dir):
            if not exec_command_with_retry("git fetch origin master", 0, MAX_RETRY_GITHUB):
                raise PrepareEnvFail

            if branch != "master":
                if not exec_command_with_retry("git config --global user.email 'you@example.com'", 0, MAX_RETRY_GITHUB):
                    raise PrepareEnvFail

                if not exec_command_with_retry("git config --global user.name 'Your Name'", 0, MAX_RETRY_GITHUB):
                    raise PrepareEnvFail

                if not exec_command_with_retry("git merge --no-ff --no-edit FETCH_HEAD", 0, MAX_RETRY_GITHUB):
                    raise PrepareEnvFail

                local("rm -rf {0}/.git".format(unique_path_dir))

        with lcd("/tmp"):
            if not exec_command_with_retry("tar -czf {0}.tar.gz {0}".format(unique_name_dir), 0, MAX_RETRY_GITHUB):
                raise PrepareEnvFail

        logger.info("The test env is ready!")

    except NetworkError as exc:
        logger.error(exc)
        raise
    except PrepareEnvFail:
        local("rm -rf {0}*".format(unique_path_dir))
        raise

    return unique_path_dir


def get_config(package_name, test_env_path, full=False):
    """Retrieve the test configuration

    Args:
        package_name: Name of the package for which to apply ``full=False``
        test_env_path: the location of the cloned test environment.
        full (boolean): See ``Returns`` section

    Returns:
        if ``full`` True, return the complete config.json file.
        Otherwise, return a dict with keys: ``['pkg_name', 'pkg_type', 'test_type']``
    """
    pkg_config = {}
    logger = Log()

    with lcd('{0}'.format(test_env_path)):
        config = json.loads(local('cat config.json', capture=True))

    logger.info('Parsing configuration')
    if not full:
        for package in config['packages']:
                if package['pkg_name'] == package_name:
                    pkg_config = package
                    break
    else:
        pkg_config = config

    logger.info('Conf is ok and has keys {0}'.format(pkg_config.keys()))
    return pkg_config



class CIJobFail(Exception):
    """Generic CI fail"""


class PrepareEnvFail(Exception):
    """Exception raised by a fail during environment preparation"""


class CITestFail(Exception):
    """Exception raised by a a fail during tests"""


class CITestUnknown(Exception):
    """Exception raised when the test is unknown"""


class CIBuildPackageFail(Exception):
    """Exception raised when the build fails"""


class CIPackageUnknown(Exception):
    """Exception raised when the package is unknown"""


class HTTPErrorCode(Exception):
    """Exception raised by a failed call on Failed HTTP Requests"""


class GitHubAPIFail(Exception):
    """Exception raised by a failed call on Github API"""


class JobIdUnknown(Exception):
    """Exception raised by a job id unknown"""


class PullRequestCollision(Exception):
    """Exception raised by a failed call on Github API"""


class FailDrawDepGraph(Exception):
    """Exception raised by a failed call on internal gplot function"""


class ForbiddenTimelineKey(Exception):
    """Exception raised by conflict key in complex plugin"""


class SimplePluginErrorConfiguration(Exception):
    """Exception raised by configuration error in simple plugin"""


class ImportPluginError(Exception):
    """Exception raised by if the plugin doesn't exist"""


class UnknownStorageMethod(Exception):
    """Exception raised by if the storage method doesn't exist"""


class WheelPackageFail(Exception):
    """Exception raised when the python wheel packaging fail"""



import base64
import pickle


def serialize(item):
    """Serialize an item

    Args:
        item (object): Python object

    Returns:
        str: The item serialized as a string
    """
    return base64.b64encode(pickle.dumps(item)).replace("=", "A")


def unserialize(item):
    """Unserialize an item

    Args:
        item (str): String to decode

    Returns:
        object: Python object
    """
    return pickle.loads(base64.b64decode(item))



import os
import pickle
import redis


class Job(object):
    def __init__(self):
        self.host = os.environ.get("LOKTAR_JOB_DB_URI", "localhost")
        self.port = os.environ.get("LOKTAR_JOB_DB_PORT", 6379)
        self.db = os.environ.get("LOKTAR_JOB_DB_NAME", 0)
        self.expire = os.environ.get("LOKTAR_JOB_EXPIRE", 86400)
        self._db = redis.StrictRedis(host=self.host, port=self.port, db=self.db)
        self._pipeline = self._db.pipeline()

    def set_job(self, job_id, job_payload):
        """Save job information in a db

        Args:
            job_id (str): the id of the current job
            job_payload (dict): the payload of the job who is store in the db

        Return:
             boolean, if the set on the db is ok True or if it failed False
        """
        return self._db.set(job_id, pickle.dumps(job_payload), ex=self.expire)

    def get_job(self, job_id):
        """Get a specific job payload

        Args:
            job_id (str): the id of the current job

        Return:
             dict, return the job payload
        """
        return pickle.loads(self._db.get(job_id))

    def get_jobs(self):
        """Get information on all jobs

        Return:
             list, who represents all payload for current job
        """
        for job_id in self._db.scan_iter():
            self._pipeline.get(job_id)

        return map(pickle.loads, self._pipeline.execute())



import datetime
import sys


class Log(object):
    def __init__(self):
        self.color = {
            "warning": "\033[93m",
            "error": "\033[91m",
            "end": "\033[0m",
        }

    def info(self, msg):
        self._print_console(msg, "info")

    def error(self, msg):
        self._print_console(msg, "error")

    def warning(self, msg):
        self._print_console(msg, "warning")

    def _print_console(self, msg, log_level):
        """Print a regular message to console"""
        if log_level == 'info':
            begin = ''
            end = ''
        elif log_level == 'error':
            begin = self.color["error"]
            end = self.color["end"]
        else:
            begin = self.color["warning"]
            end = self.color["end"]

        sys.stdout.write("{0}[{1}] {2}{3}\n".format(begin,
                                                    datetime.datetime.now(),
                                                    msg,
                                                    end))
        sys.stdout.flush()



from github import Github as GitHub
import requests
import StringIO

from loktar.decorators import retry
from loktar.environment import GITHUB_INFO
from loktar.log import Log

logger = Log()


class Github(object):
    """Wrapper for the github3 library

    Args:
        login (str): login for GitHub
        password (str): password for GitHub
        github_organization (str): this is the github organization for get back the repository, default value None.
                                   Also can be set by environment variable LOKTAR_GITHUB_INFO_ORGANIZATION
        github_repository (str): this is the target repository to download, default value None
                                 Also can be set by environment variable LOKTAR_GITHUB_INFO_REPOSITORY

    """

    def __init__(self, login, password, github_organization=None, github_repository=None):
        self._connexion = GitHub(login, password)
        self.organization_name = GITHUB_INFO["repository"] if github_organization is None else github_organization
        self.repository_name = GITHUB_INFO["repository"] if github_repository is None else github_repository
        self._repository = self._connexion.get_organization(self.organization_name).get_repo(self.repository_name)
        # Used for cache
        self.pull_requests_cache = {}
        self.logger = Log()

    @retry
    def search_pull_request_id(self, branch):
        """Look for an open pull request id given a branch

        Args:
            branch (str): name of the branch

        Returns:
            int or None: the id of the pull request or None
        """
        pr_id = None
        for pr in self.get_pull_requests(state='open'):
            self.logger.info('Looking at branch {0}'.format(pr.head.ref))
            if pr.head.ref == branch:
                self.logger.info('Found pull request {0} with id {1}'.format(pr, pr.number))
                pr_id = pr.number
                break
        return pr_id

    @retry
    def get_pull_requests(self, state='open', use_cache=False):
        """Get a pull request from the GitHub API

        Args:
            state (str): State of the pull request
            use_cache (Optional[bool]): If True, only return cached pull requests. Otherwise, make another request.

        Returns:
            a list of PullRequest instance
        """
        if use_cache and self.pull_requests_cache:
            pull_requests = self.pull_requests_cache
        else:
            pull_requests = self._repository.get_pulls(state=state)
            self._cache_pull_requests(pull_requests)
        return pull_requests

    def _cache_pull_requests(self, pull_requests=None):
        """Cache a list of pull requests

        Args:
            pull_requests (list of github.pullrequest.PullRequest)
        """
        for pull_request in pull_requests:
            self.pull_requests_cache[pull_request.number] = pull_request

    @retry
    def get_pull_request(self, pull_request_id, use_cache=True):
        """Get a pull request from the GitHub API

        Args:
            pull_request_id (int): ID of the pull request
            use_cache (Optional[bool]): If True, only return cached pull requests. Otherwise, make another request.
                If we cannot find the pull request in the cache, we make another request.

        Returns:
            a PullRequest instance
        """
        if use_cache and pull_request_id in self.pull_requests_cache:
            pull_request = self.pull_requests_cache[pull_request_id]
        else:
            pull_request = self._repository.get_pull(pull_request_id)
            self._cache_pull_requests([pull_request])
        return pull_request

    @retry
    def get_commit_message_modified_files_on_pull_request(self, pull_request_id):
        """Retrieve the commit messages from the pull request and associate the modified files for it.

        Args:
            pull_request_id (int): ID of the pull request

        Returns:
            dict of str: list: Comments of the commits linked to the pull request are keys, modified files are values.
        """
        pr = self.get_pull_request(pull_request_id)
        commits = pr.get_commits()
        # Return message: files
        dict_message_files = {}
        for commit in commits:
            dict_message_files.setdefault(commit.commit.message, [])
            dict_message_files[commit.commit.message].extend(map(lambda file_: file_.filename, commit.files))
        return dict_message_files

    @retry
    def get_git_branch_from_pull_request(self, pull_request_id):
        """Retrieve the branch name from the pull request id

        Args:
            pull_request_id (int): ID of the pull request

        Returns:
            A branch name
        """
        pr = self.get_pull_request(pull_request_id)
        return pr.head.ref

    @retry
    def get_last_statuses_from_pull_request(self, pull_request_id, exclude_head=True):
        """Get the statuses from a pull request ID.

        Args:
            pull_request_id (int): ID of the pull request
            exclude_head (bool): If True, skip the head commit

        Returns:
            a tuple of github.Commit, list of github.CommitStatus.CommitStatus: List of statuses.
                Empty if no status was found.
        """
        pr = self.get_pull_request(pull_request_id)
        for commit in pr.get_commits().reversed:
            if exclude_head and commit.sha == pr.head.sha:
                continue
            statuses = list(commit.get_statuses())
            if statuses:
                return commit, statuses
        return None, []

    @retry
    def get_last_commits_from_pull_request(self, pull_request_id, until_commit=None):
        """Get the commit from a pull request ID.

        Args:
            pull_request_id (int): ID of the pull request
            until_commit (github.Commit.Commit): Stop at a specific commit (will be excluded)

        Returns:
            list of github.Commit.Commit: the last commits
        """
        commits = []

        pr = self.get_pull_request(pull_request_id)
        for commit in pr.get_commits().reversed:
            if until_commit is not None and commit.sha == until_commit.sha:
                break
            commits.append(commit)
        return commits

    @retry
    def create_pull_request_comment(self, pull_request_id, comment, check_unique=False):
        """Create an issue comment on the pull request

        Args:
            pull_request_id (int): ID of the pull request
            comment (basestring): Comment text to post
            check_unique (bool): If True do not comment if it has already been posted.

        Returns:
            github.IssueComment.IssueComment: The comment that was created
        """
        pr = self.get_pull_request(pull_request_id)
        if check_unique:
            issue_comments = pr.get_issue_comments()
            for issue_comment in issue_comments:
                if issue_comment.body == comment:
                    return
        return pr.create_issue_comment(comment)


@retry
def fetch_github_file(url, token):
    """Fetch a file from GitHub

    Args:
        url (str): URL the file is at.
        token (str): Authorization token for GitHub.

    Returns:
        str: the raw content of the file.
    """
    logger = Log()
    headers = {
        'Authorization': 'token {0}'.format(str(token)),
        'Accept': 'application/vnd.github.v3.raw',
    }
    logger.info('Fetching file at {0}'.format(url))
    req = requests.get(url, headers=headers, allow_redirects=True)
    m_file = StringIO.StringIO()
    m_file.write(req.content)
    return m_file.getvalue()



from loktar.environment import PLUGINS_LOCATIONS
from loktar.exceptions import CITestFail
from loktar.exceptions import CIBuildPackageFail
from loktar.exceptions import ImportPluginError
from loktar.log import Log
from loktar.plugin import find_plugin


def strategy_runner(package, run_type):
    """Run the packaging functions

    Args:
        package (dict): package_config
        run_type (str): Represent the strategy to run on a package (test or artifact)

    Raises:
        CITestFail: some error occurred during the test
        CITestUnknown: wrong value for config['test_type']
        CIBuildPackageFail: some error occurred during a packaging
        ImportPluginError: Fail to find / import a plugin
    """

    if run_type not in ["test", "build"]:
        raise ValueError("run_type must be equal to 'test' or 'build', actual value: {0}".format(run_type))

    logger = Log()
    params = {"type": "test_type", "exception": CITestFail}\
        if run_type == "test" else {"type": "pkg_type", "exception": CIBuildPackageFail}
    plugins_location = PLUGINS_LOCATIONS.split(",") if type(PLUGINS_LOCATIONS) is str else PLUGINS_LOCATIONS
    try:
        runner = find_plugin(package[params["type"]], plugins_location)

    except ImportPluginError:
        raise

    try:
        runner.run(package)
    except Exception as e:
        logger.error(repr(e))
        raise params["exception"]




from fabric.api import cd
from fabric.api import lcd
from fabric.api import settings
import importlib
import sys

from loktar.cmd import exe
from loktar.environment import ROOT_PATH
from loktar.exceptions import CITestFail
from loktar.exceptions import ForbiddenTimelineKey
from loktar.exceptions import ImportPluginError
from loktar.exceptions import SimplePluginErrorConfiguration
from loktar.log import Log


class SimplePlugin(object):
    def __init__(self, package_info, config, remote=False):
        """Constructor for the SimplePlugin

        Args:
            package_info (dict): represent all informations for the target package in the config.json
            config (dict): this is the configuration plugin. It contains 2 keys 'run' & 'clean'.
            remote (bool): Indicate if it has to execute commands in remote or not

        Raise:
            SimplePluginErrorConfiguration: An error occurred when a key is missing in the configuration
        """
        self.logger = Log()
        self.config = config
        self.package_info = package_info
        self.remote = remote
        self.cwd = lcd if self.remote is False else cd

        try:
            assert "package_location" in package_info
        except AssertionError:
            self.path = "{0}/{1}/{2}".format(ROOT_PATH["container"],
                                             package_info["pkg_dir"],
                                             package_info["pkg_name"]) \
                                             if "pkg_dir" in package_info else \
                                             "{0}/{1}".format(ROOT_PATH["container"],
                                                              package_info["pkg_name"])

        try:
            self.cmd = self.config["command"]
        except KeyError:
            raise SimplePluginErrorConfiguration()

    def __command(self, cmd):
        """Run the command indicated in the yaml file in the package directory

        Raise:
            CITestFail: An error occurred when the test failed
        """

        with self.cwd(self.package_info["package_location"]):
            with settings(warn_only=True):
                if not exe(cmd, remote=self.remote):
                    raise CITestFail("Test failed")

    def _base_run(self):
        try:
            self.__command(self.cmd["run"])
        except CITestFail:
            self._base_clean()
            raise

    def _base_clean(self):
        self.__command(self.cmd["clean"])


class ComplexPlugin(SimplePlugin):
    def __init__(self, package_info, config):
        """Constructor for the ComplexPlugin, child of SimplePlugin

        Args:
            package_info (dict): represent all informations for the target package in the config.json
            config (dict): this is the configuration plugin. It contains 2 keys 'run' & 'clean'.

        Raise:
            SimplePluginErrorConfiguration: An error occurred when a key is missing in the configuration
        """
        SimplePlugin.__init__(self, package_info, config)
        self.timeline = dict()
        self.share_memory = dict()
        self.__origin = {
            50: self._base_run,
            95: self._base_clean
        }

    def _run(self):
        """Run the timeline
        Args:
            timeline (dict): represent the user timeline, it will be merged with the origin timeline

        Raise:
            ForbiddenTimelineKey: An error occurred when the plugin try to used a reserved key
        """
        try:
            assert not (set(self.__origin.keys()) & set(self.timeline))
        except AssertionError:
            raise ForbiddenTimelineKey("Timeline key: 50 & 95 are reserved")

        self.timeline.update(self.__origin)

        for ref in sorted(self.timeline):
            self.timeline[ref]()


def find_plugin(plugin_name, plugin_locations):
    """
    Args:
        plugin_name (str): the plugin to search
        plugin_locations (list): locations of plugins

    Raises:
        ImportPluginError: it raise if the plugin cannot import or the plugin is not found

    Returns
        module (module): return the plugin
    """
    sys.path += plugin_locations
    errors = list()
    logger = Log()

    try:
        return importlib.import_module("loktar.plugins.{0}".format(plugin_name))
    except ImportError as e:
        errors.append(str(e))

    try:
        return importlib.import_module(plugin_name)
    except ImportError as e:
        errors.append(str(e))
        logger.error("\n".join(errors))
        raise ImportPluginError



import copy
from fabric.api import lcd
from fabric.api import local
from itertools import product
import json
import networkx as nx
import numpy as np
import os
import re
import time
import urllib

from loktar.decorators import retry
from loktar.environment import GITHUB_INFO
from loktar.environment import GITHUB_TOKEN
from loktar.exceptions import CIJobFail
from loktar.exceptions import FailDrawDepGraph
from loktar.log import Log
from loktar.parser import parse_statuses
from loktar.scm import fetch_github_file
from loktar.scm import Github

logger = Log()


def docker_deps(basepath):
    """Get docker image dependencies.

    Args:
        basepath (str): Path to the package

    Returns:
        set: Dependencies names.
    """
    docker_file = os.path.join(basepath, 'Dockerfile')
    if os.path.exists(docker_file):
        with open(docker_file, 'r') as f:
            dockerfile = f.read()
        re_dockerfile = re.compile('FROM (.+)')
        docker_from = re_dockerfile.findall(dockerfile)
        assert len(docker_from) <= 1
        if docker_from:
            docker_from = docker_from[0]
            docker_image_name_tag = docker_from.rsplit('/', 1)[-1]
            docker_image_name = docker_image_name_tag.split(':', 1)[0]
            return {docker_image_name}
        else:
            return set()
    else:
        return set()


def python_deps(basepath):
    """Get python dependencies from requirements.txt and test_requirements.txt.

    Args:
        basepath (str): Path to the package

    Returns:
        set: Requirements names.
    """
    requirements = []

    requirements_file = os.path.join(basepath, 'requirements.txt')
    test_requirements_file = os.path.join(basepath, 'test_requirements.txt')
    app_requirements_file = os.path.join(basepath, 'app', 'requirements.txt')
    app_test_requirements_file = os.path.join(basepath, 'app', 'test_requirements.txt')

    for req_file in (requirements_file, test_requirements_file, app_requirements_file, app_test_requirements_file):
        if os.path.exists(req_file):
            with open(req_file, 'r') as f:
                requirements.extend(f.readlines())

    requirements_names = set(filter(None, map(lambda x: re.split('==|>=|>|<=|<', x)[0], requirements)))
    return requirements_names


def deps(basepath):
    """Gather all dependencies for this package

    Args:
        basepath (str): Path to the package

    Returns:
        set: Dependencies names.
    """
    return python_deps(basepath) | docker_deps(basepath)


def get_excluded_deps(packages, dict_message_files, modified_packages):
    """Compute which package's dependencies should be ignored based on commit messages.

    This does not take into account modified dependencies of an package whose dependencies are excluded.
    We only need to know if an packages non-modified dependencies should be ignored or not, the rest is
    done in ``dependency_graph_from_modified_packages``.

    Args:
        packages (dict): all packages. Keys are packages names, values are directly taken from config.json
        dict_message_files (dict of str: list): Comments of the commits linked to the pull request are keys,
            modified files are values. If empty, the returned set is empty.
        modified_packages (set of str): names of the packages that were modified

    Returns:
        set of str: items are packages names whose dependencies must not be added to the dependency graph

    Examples:
        >>> get_excluded_deps({'package': {'exclude_dependencies_only_on_keywords': ['CLN', 'BLD']}},
        ... {'CLN/BLD: Cleaned some files. I do not care about dependencies': 'package/modified.py'},
        ... {'package'})
        {'package'}

        >>> get_excluded_deps({'package': {'exclude_dependencies_only_on_keywords': ['CLN', 'BLD']}},
        ... {'BUG: Fixed a serious bug! This should rebuild dependencies': 'package/modified.py'},
        ... {})
        {'package'}
    """

    if not dict_message_files:
        return set()

    # Packages whose dependencies should be ignored
    # Initialize to all of them and delete them one by one if some keywords are not exclusion keywords.
    exclude_dep = set(copy.deepcopy(modified_packages))

    for commit_message, modified_files in dict_message_files.items():
        for path in modified_files:
            package_name = package_from_path(path, packages)
            if package_name is None:
                continue
            # If we have set an exclude condition and this condition is matched
            elif 'exclude_dependencies_only_on_keywords' in packages[package_name]:
                keywords = {key.strip() for key in commit_message.split(':')[0].split('/')}
                keywords_exclude = set(packages[package_name]['exclude_dependencies_only_on_keywords'])
                # If there is a keyword that is not excluded
                if keywords - keywords_exclude:
                    exclude_dep -= {package_name}
            else:
                exclude_dep -= {package_name}
    return exclude_dep


def package_from_path(path, packages):
    """Get the name of an package from a path.

    Args:
        path: relative path starting with the potential package name (Example: serializers/http/)
        packages (dict): all packages. Keys are packages names, values are directly taken from config.json

    Returns:
        the package name or None if the path is not inside an package
    """
    for package_name, config in packages.iteritems():
        if 'pkg_dir' in config and path.startswith(os.path.join(config['pkg_dir'], package_name)):
            return package_name
        elif path.startswith(package_name):
            return package_name


def package_path(package):
    """Get the path of a package

    Args:
        package (dict): Package configuration

    Returns:
        str: The package path relative to the repository root.
    """
    if 'pkg_dir' in package:
        return os.path.join(package['pkg_dir'], package['pkg_name'])
    else:
        return package['pkg_name']


def graph_edges(packages_requirements, modified_packages, exclude_dep_from_packages):
        """Recursively find graph edges

        Args:
            packages_requirements (dict of list): Dictionary of package_name: requirements
            modified_packages (set): Package names that are considered modified
            exclude_dep_from_packages (set): Do not analyze the dependencies for the packages in this list

        Returns:
            list of tuple: Graph edges between packages.
        """
        if not modified_packages:
            return []

        edges_list = []
        next_modified_packages = set()
        packages_done = set()

        for package_name in packages_requirements:
            requirements = packages_requirements[package_name]
            # but whose dependencies are not excluded (iff this was not a modified package)
            modified_requirements = (requirements & modified_packages - exclude_dep_from_packages
                                     if package_name not in modified_packages
                                     else requirements & modified_packages)

            if modified_requirements:
                # Add the edges from the modified requirements towards this package
                edges = map(lambda x: (x, package_name), modified_requirements)
                edges_list.extend(edges)
                packages_done |= modified_requirements
                next_modified_packages |= {package_name}

        for package_name in packages_done:
            packages_requirements.pop(package_name, None)

        edges_list.extend(graph_edges(packages_requirements, next_modified_packages, exclude_dep_from_packages))
        return edges_list


def dependency_graph_from_modified_packages(repo_path,
                                            packages,
                                            modified_packages,
                                            exclude_dep_from_packages=None):
    """Parse requirement with a pattern for generating a dependencies graph

    Args:
        repo_path (str): this is the path where the repo to parse is
        packages (dict): all packages. Keys are packages names, values are directly taken from config.json
        modified_packages (set): This is a set a modified package
        exclude_dep_from_packages (Optional[set]): Do not analyze the dependencies for the packages in this list.
            Defaults to None.

    Returns:
        networkx.classes.digraph.DiGraph: The dependency graph.
    """
    if exclude_dep_from_packages is None:
        exclude_dep_from_packages = set()

    modified_packages = set(modified_packages)
    packages_requirements = {}

    # Go through all the packages
    for package_name in packages:
        requirements = get_package_requirements(package_name, packages, repo_path)
        if requirements:
            packages_requirements[package_name] = requirements

    edges_list = graph_edges(packages_requirements, modified_packages, exclude_dep_from_packages)

    directed_graph = nx.DiGraph(edges_list)
    directed_graph.add_nodes_from(modified_packages)

    return directed_graph


def get_package_requirements(package_name, packages, repo_path, restrict_requirements_types=None):
    """Get the requirements of a package

    Args:
        package_name (str): Name of the package to consider
        packages (dict): All packages. Keys are packages names, values are directly taken from config.json
        repo_path (str): This is the path where the repo to parse is
        restrict_requirements_types (Optional[list or None]): Restrict the requirements type to some type.
            Available types are: celery-microservice, library, swagger-rest-microservice, skeleton, docker.
            Defaults to None.

    Returns:
        set of str: List of requirements for this package
    """

    build_deps_path = os.path.join(repo_path, package_path(packages[package_name]))
    requirements = deps(build_deps_path) & set(packages.keys())

    if restrict_requirements_types is not None:
        packages_restricted = filter(lambda pkg: pkg.get('type') in restrict_requirements_types, packages.values())
        packages_restricted = set(map(lambda pkg: pkg['pkg_name'], packages_restricted))
        requirements = requirements & packages_restricted

    if not requirements:
        logger.info('No internal requirements found for package {0} in {1}'.format(package_name, build_deps_path))
    return requirements


def output_levels(graph):
    """Generate relationship level dependencies with a directed graph

    Args:
        graph (networkx.classes.digraph.DiGraph): Represent a directed graph

    Return:
        Return a list of dependencies level, top level to the lower level
    """
    top_sort = list(nx.topological_sort_recursive(graph))

    levels = [[] for _ in xrange(len(top_sort))]

    for level, node in enumerate(top_sort):
        potential_level = level
        # While we are not connected to this level we push the package down a level
        while all((level_node, node) not in graph.edges() for level_node in levels[potential_level]) \
                and potential_level >= 0:
            potential_level -= 1

        # Potential_level is now below the one where the node should go (where we are not connected)
        potential_level += 1
        levels[potential_level].append(node)
    levels = filter(None, levels)
    return levels


def get_directed_components(graph):
    """Generate a list a directed graph

    Args:
        graph (networkx.classes.digraph.DiGraph): Represent a directed graph

    Returns:
        List of networkx.classes.digraph.DiGraph
    """
    undirected_connected_components = list(nx.connected_component_subgraphs(graph.to_undirected()))
    directed_connected_components = [graph.subgraph(comp_G.nodes()) for comp_G in undirected_connected_components]
    return directed_connected_components


def dependencies_layout(dependencies_levels, scale=1):
    """Generate a layout for the dependencies graph

    Args:
        dependencies_levels: A list of list with dependencies levels
        scale (float): Scale factor for positions

    Returns:
        dict: A dictionary of positions keyed by node
    """
    pos_x = np.empty(shape=0)
    pos_y = np.empty(shape=0)

    nodes = []

    scale *= 3

    component_top = 0
    for component_i, component in enumerate(dependencies_levels):
        pos_packages_x = np.empty(shape=0)
        pos_packages_y = np.empty(shape=0)
        for level_i, level in enumerate(component):
            nodes.extend(level)
            pos_packages_x = np.concatenate((pos_packages_x, np.zeros(len(level)) + level_i * scale))
            pos_packages_y = np.concatenate((pos_packages_y,
                                             np.linspace(0, scale * len(level), len(level)) - scale * len(level) / 2.))
        # We put this component on top of the previous one, adding a 'scale' margin
        pos_packages_y += 3 * scale + component_top
        # We update the total height
        component_top = pos_packages_y.max()

        pos_x = np.concatenate((pos_x, pos_packages_x))
        pos_y = np.concatenate((pos_y, pos_packages_y))

    pos = np.vstack((pos_x, pos_y))

    # Since the layout has grown arbitrarily big, we rescale it and center it
    pos = dict(zip(nodes, pos.T))
    return pos


def gplot(graph, dependencies_levels, save_path=None, output_type=None):
    """Draw a the dependency graph

    Args:
        graph (networkx.classes.digraph.DiGraph): Represent a directed graph
        dependencies_levels: A list of list with dependencies levels
        save_path (str): The path where the file is saved
        output_type (str): The output file can be a dot file or a png
    """
    name_file = str(time.time()).replace(".", "_")

    final_path = "{0}/{1}.{2}".format(save_path, name_file, output_type)

    if output_type is not None and save_path is not None:
        logger.info("Storing {0}.{1} in {2}".format(name_file, output_type, save_path))

    if output_type == "png":
        import matplotlib
        matplotlib.use('Agg')
        import matplotlib.pyplot as plt
        logger.info("Drawing the graph")
        plt.figure(figsize=(12, 14))
        nx.draw_networkx(graph, pos=dependencies_layout(dependencies_levels), edge_color='g', node_size=3000)
        plt.savefig(final_path)
        logger.info("The graph is drawn")
    elif output_type == "dot":
        logger.info("Generating dot file")
        nx.write_dot(graph, final_path)
        logger.info("The dot file is generated")
    elif output_type is None:
        logger.info("output type is {0}".format(output_type))
    else:
        logger.error("The output type is unknown")
        raise FailDrawDepGraph

    logger.info("The file {0} is stored".format(final_path))
    return name_file


def gen_dependencies_level(directed_graph, draw=False):
    """Generate dependencies levels

    Args:
        directed_graph (networkx.classes.digraph.DiGraph): Graph that represents relationship dependencies
        draw (Boolean): of the function generate an image of dependencies level

    Returns:
        a tuple of
          - a boolean: True or False
          - a list of list with dependencies levels
    """
    name_file = None
    try:
        assert nx.is_directed_acyclic_graph(directed_graph)
    except AssertionError:
        logger.error('A cycle has been detected in the graph.')
        return False, None,

    logger.info('Generating the dependency levels')

    dependencies_levels = [output_levels(comp) for comp in get_directed_components(directed_graph)]
    logger.info('The dependency levels have been generated')
    if draw:
        name_file = gplot(directed_graph, dependencies_levels, save_path='/home/pwned/ci_img', output_type='png')
        with lcd("/home/pwned/ci_img"):
            local("git checkout master")
            local("git fetch origin")
            local("git merge origin/master")
            local("git add .")
            local("git commit -m 'Add {0}'".format(name_file))
            local("git push origin")

    return True, dependencies_levels, name_file


def get_do_not_touch_packages(id_pr, packages, scm, dep_graph, rebuild=False):
    """Get packages that should not be touched (except if one of their dependencies was modified)

    These are packages which have only green builds, that were not modified in the last commits, and whose
    relationships between each other involve _only_ themselves.

    Args:
        id_pr: ID of the pull request
        packages (dict): all packages. Keys are packages names, values are directly taken from config.json
        scm (loktar_api.scm.Github): Internal Github instance
        dep_graph (networkx.classes.digraph.DiGraph): Graph that represents relationships between dependencies
        rebuild (bool): If True (default), skip the head commit

    Returns:
        set: Set of packages that should not be touched
    """

    if rebuild:  # If rebuild is True, we only consider the last commit's statuses
        last_status_commit, statuses = scm.get_last_statuses_from_pull_request(id_pr, exclude_head=False)
        pr = scm.get_pull_request(id_pr)
        if last_status_commit.sha == pr.head.sha:
            modified_packages_last_commits = set()
        else:
            raise CIJobFail('Cannot rebuild if no build was launched in the first place.')

    else:  # Otherwise, we get the last commit that has statuses before the last commit of the pull request
        last_status_commit, statuses = scm.get_last_statuses_from_pull_request(id_pr)
        last_commits = scm.get_last_commits_from_pull_request(id_pr, until_commit=last_status_commit)
        # Get the modified package in the last commit
        all_commit_files = [commit_file for commit in last_commits for commit_file in commit.files]
        modified_files = map(lambda file_: file_.filename, all_commit_files)
        modified_packages_last_commits = {package_from_path(path, packages) for path in modified_files}
        modified_packages_last_commits -= {None}

    green_builds, red_builds = parse_statuses(statuses)

    logger.info('Packages with green builds: {0}'.format(green_builds))
    logger.info('Packages with red builds: {0}'.format(red_builds))
    logger.info('Packages modified in the last commit: {0}'.format(modified_packages_last_commits))

    # Packages that should not be touched are package that got completely green builds
    # and that were not modified during the last commit
    do_not_touch_packages = green_builds - red_builds - modified_packages_last_commits

    if do_not_touch_packages:
        # In the following algorithm we use an augmented dep graph where we add a node which points to every other
        # node, and this node is added temporarily to do_not_touch_packages
        # This is used to avoid the case where the root(s) of the graph (set of nodes that are topologically less
        # than every other node) are not modified, then creating a path where there is an package not in
        # do_not_touch_packages and solving the corner case of the path algorithm.
        null_node = '__null__'
        augmented_dep_graph = copy.deepcopy(dep_graph)
        augmented_dep_graph.add_edges_from([(null_node, node) for node in augmented_dep_graph.nodes()])
        do_not_touch_packages |= {null_node}

        # TODO(Simplify the path check. Some packages are checked several times)

        # Now we check that for every path between packages in this list a path between them only contains packages
        # that should not be touched. If this is not the case we add the target package to have_path
        have_incomplete_path = set()
        for package_from, package_to in product(do_not_touch_packages, do_not_touch_packages):
            paths = nx.all_simple_paths(augmented_dep_graph, package_from, package_to)
            # We remove every 2-path containing the null node
            paths = filter(lambda path: not (len(path) == 2 and null_node in path), paths)
            for path in paths:
                # If there are packages in the path that are not in do_not_touch, add the target package to the set
                if set(path) - do_not_touch_packages:
                    have_incomplete_path |= {package_to}

        do_not_touch_packages -= have_incomplete_path
        do_not_touch_packages -= {null_node}

        logger.info('Packages that have an incomplete path directed to them: {0}'.format(have_incomplete_path))
    logger.info('Packages that should not be touched: {0}'.format(do_not_touch_packages))
    return do_not_touch_packages


@retry
def pull_request_information(id_pr, workspace):
    """Relevant information for packages from a pull request

    Args:
        id_pr: ID of the pull request
        workspace:

    Returns:
        tuple
    """
    scm = Github(GITHUB_INFO["login"]["user"], GITHUB_INFO["login"]["password"])

    if id_pr is not None:
        dict_message_files = scm.get_commit_message_modified_files_on_pull_request(id_pr)
        # Filter out Merge remote-tracking commits and Merge branch
        dict_message_files = {key: value for key, value in dict_message_files.iteritems()
                              if (not key.startswith('Merge remote-tracking branch') and
                                  not key.startswith('Merge branch'))}
        git_branch = scm.get_git_branch_from_pull_request(id_pr)
        config = json.loads(fetch_github_file('https://raw.githubusercontent.com/{0}/{1}/{2}/config.json'
                                              .format(GITHUB_INFO["organization"],
                                                      GITHUB_INFO["repository"],
                                                      urllib.quote(git_branch)),
                                              GITHUB_TOKEN))
        map_pkg = config["packages"]
        # Concatenate the lists that are values of dict_message_files
        git_diff = [item for _list in dict_message_files.values() for item in _list]
    else:
        logger.error("Pull request not found: #{0}".format(id_pr))
        raise ValueError('No PR #{0}'.format(id_pr))

    packages = {pkg["pkg_name"]: pkg for pkg in map_pkg}
    modified_packages = {package_from_path(path, packages) for path in git_diff} - {None}
    exclude_dep = get_excluded_deps(packages, dict_message_files, modified_packages)
    logger.info('The following packages\' dependencies will be ignored if unmodified: {0}'
                .format(exclude_dep))

    with lcd(os.path.join(workspace)):
        local("git checkout master")
        local("git fetch origin")
        local("git merge origin/master")
        local("git checkout {0}".format(git_branch))
        local("git merge origin/{}".format(git_branch))
        dep_graph = dependency_graph_from_modified_packages(workspace,
                                                            packages,
                                                            modified_packages,
                                                            exclude_dep)
    return config, dep_graph, modified_packages



from loktar.job import context_to_build_params


def parse_statuses(statuses):
    """Parse a list of statuses

    Some packages can be simulteneouly in green builds and red builds, if they were built with
    more than one different build type.

    Args:
        statuses (iterator of github.CommitStatus.CommitStatus): List of commit statuses

    Returns:
        a tuple of sets: green_builds, red_builds
    """
    green_builds = set()
    red_builds = set()
    for status in statuses:
        context = status.raw_data['context']
        try:
            package, type_build = context_to_build_params(context)
        except ValueError:
            continue
        if status.state == 'success':
            green_builds |= {(package, type_build)}
        else:
            red_builds |= {(package, type_build)}

    # If a tuple (package, type_build) is green and also appears in red, remove it from red
    # (since red are also pending)
    red_builds -= green_builds

    green_builds = {package for package, _ in green_builds}
    red_builds = {package for package, _ in red_builds}

    return green_builds, red_builds



from loktar.plugin import ComplexPlugin


def run(*args, **kwargs):
    Make(args[0]).run()


class Make(ComplexPlugin):
        def __init__(self, package_info):
            ComplexPlugin.__init__(self, package_info,
                                   {
                                       "command": {
                                           "run": "make ci",
                                           "clean": "make clean"
                                       }
                                   })
            self.timeline = {
                60: self.get_results,
            }

        def run(self):
            self._run()

        def get_results(self):
            print 'ToDo'




from loktar.environment import PLUGINS_LOCATIONS
from loktar.exceptions import CITestFail
from loktar.exceptions import ImportPluginError
from loktar.log import Log
from loktar.plugin import find_plugin


def run_test(package):
    """Run the packaging functions

    Args:
        package (dict): package_config

    Raises:
        CITestFail: some error occured during the test
        CITestUnknown: wrong value for config['test_type']
    """
    logger = Log()
    plugins_location = PLUGINS_LOCATIONS.split(",") if type(PLUGINS_LOCATIONS) is str else PLUGINS_LOCATIONS
    try:
        # test_runner = eval("{0}.{1}".format(find_plugin(package["test_type"], plugins_location),
        #                                    package["test_type"].title().replace("_", "")))
        test_runner = find_plugin(package["test_type"], plugins_location)

    except ImportPluginError:
        raise

    try:
        test_runner.run(package)
    except Exception as e:
        logger.error(repr(e))
        raise CITestFail




import os
import sys

from fabric.api import local

from loktar.notifications import define_job_status_on_github_commit


def exit_properly(exit_code,
                  commit_id=None,
                  pre_command_exit=None,
                  notif_msg=None,
                  description=None,
                  **kwargs):
    """Exist after notifying github

    Args:
        exit_code:
        commit_id:
        pre_command_exit:
        notif_msg:
        description:

    """
    if commit_id is not None:
        define_job_status_on_github_commit(commit_id,
                                           'success' if exit_code == 0 else 'error',
                                           os.environ.get('BUILD_URL') + 'console#footer-container',
                                           context=notif_msg if notif_msg is not None else 'CI',
                                           description=description,
                                           user=kwargs.get("user", None),
                                           password=kwargs.get("password", None),
                                           organization=kwargs.get("organization", None),
                                           repository=kwargs.get("repository", None))
    if pre_command_exit is not None:
        local(pre_command_exit)

    sys.exit(exit_code)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import find_packages
from pip.req import parse_requirements


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = [str(i.req) for i in parse_requirements("requirements.txt", session=False)]
test_requirements = [str(i.req) for i in parse_requirements("test_requirements.txt", session=False)]

VERSION = '1'

setup(
    name='loktar',
    version=str(VERSION),
    description='ci',
    long_description=readme,
    author='Damien Goldenberg',
    author_email='damdam.gold@gmail.com',
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    keywords='ci',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



from __future__ import division

import logging
import random
import re
import time

import jenkins
from jenkinsapi.constants import STATUS_SUCCESS
from jenkinsapi.custom_exceptions import NotBuiltYet
from jenkinsapi.jenkins import Jenkins
from requests import HTTPError

from loktar.environment import GITHUB_INFO
from loktar.environment import JENKINS_PROTOCOL
from loktar.exceptions import CIJobFail
from loktar.log import Log
from loktar.notifications import define_job_status_on_github_commit
from loktar.scm import Github
from loktar.serialize import serialize

logger = logging.getLogger(__name__)

log = Log()


def ci_downstream(ci_config, pkg_name, type_task, params, job_format="{0} - {1}"):
    """Send a job to the ci

    Args:
        ci_config: part of the config.json file under the ``'jenkins'`` key
        pkg_name: name of the package to build
        type_task: apparently only ``'superman'``
        params: parameters sent to jenkins
    """
    ci_server = jenkins.Jenkins(ci_config['host'],
                                ci_config['user'],
                                ci_config['password'])
    ci_server.build_job(job_format.format(pkg_name, type_task), params)


def job_manager(ci_config,
                component,
                component_id,
                git_branch,
                commit_id,
                committer,
                test_env_path):
    """Build all the levels for a component

    Args:
        ci_config: part of the config.json file under the ``'jenkins'`` key
        component: a list of loktar.dependency levels. Example:
        [['my_firstlib'], ['microservice_depending on_my_firstlib']]
        component_id: string with component number/component name
        git_branch: git branch name to test / build
        commit_id: commit_id
        committer: name of the person which committed
        test_env_path: the path where the environment is store (eg: /tmp/toto)
    """
    types_build = ['test', 'artifact'] if git_branch != 'master' else ['artifactmaster']

    jenkins_instance = Jenkins('{0}://{1}'.format(JENKINS_PROTOCOL,
                                                  ci_config['host']),
                               username=ci_config['user'],
                               password=ci_config['password'])

    for actual_number_lvl, lvl in enumerate(component):
        for type_build in set(types_build) - {'artifactmaster'}:
            for package in lvl:

                context = build_params_to_context(package, type_build)
                define_job_status_on_github_commit(commit_id,
                                                   "pending",
                                                   'http://github.com',
                                                   context=context,
                                                   description='Build awaiting launch')

    for actual_number_lvl, lvl in enumerate(component):
        for type_build in types_build:
            log.info('Building {type_build} for component {component_id}, level {lvl}/{lvl_nb}'
                     .format(type_build=type_build,
                             component_id=component_id,
                             lvl=actual_number_lvl + 1,
                             lvl_nb=len(component)))
            launch_jobs(jenkins_instance,
                        commit_id,
                        committer,
                        git_branch,
                        lvl,
                        type_build,
                        test_env_path)


def split_on_condition(seq, condition):
    """Split an iterator on a condition

    Args:
        seq (iterator): Iterator to process
        condition (function): Function representing the condition

    Returns:
        tuple of list: first element is a list of elements for which ``condition`` is True, the second element is a list
        for which ``condition`` is False.
    """
    true_list, false_list = [], []
    for item in seq:
        (true_list if condition(item) else false_list).append(item)
    return true_list, false_list


def is_good(build):
    """Since jenkinsapi.Build.is_good is not good enough, I'm writing this one

    Args:
        build: instance of jenkinsapi.Build

    Returns:
        a boolean True or False
    """
    tries = 12
    while build._data['result'] is None and tries > 0:
        build.poll()
        time.sleep(0.5)
        tries -= 1
    return build._data['result'] == STATUS_SUCCESS


def launch_jobs(jenkins_instance,
                commit_id,
                committer,
                git_branch,
                lvl,
                type_build,
                test_env_path):

    """Launch the jobs defined through the loktar.dependency graph

    Args:
        jenkins_instance: instance of the jenkins class
        commit_id: commit_id
        committer: name of the person which committed
        git_branch: git branch name to test / build
        lvl: level being run
        type_build: 'test' or 'artifact'
        test_env_path: the location of the cloned test environment.
    """
    launched_queues = []
    for package in lvl:
        launched_queues.append(launch_queue(jenkins_instance,
                                            commit_id,
                                            committer,
                                            package,
                                            git_branch,
                                            type_build,
                                            test_env_path))

    scm = Github(GITHUB_INFO['login']['user'], GITHUB_INFO['login']['password'])
    pr_id = scm.search_pull_request_id(git_branch)

    # Get an instance of the queue
    queue_instance = jenkins_instance.get_queue()
    running_builds = []

    i = 1

    # While there are queued items and running builds
    while launched_queues + running_builds:
        # We check for the first stopping condition, which is that the Pull request received a new commit
        if pr_id is not None:
            pr = scm.get_pull_request(pr_id)
            commit_sha = pr.head.sha
            if commit_id != commit_sha:
                log.info('PR #{0} received a new commit! Stopping this build.'.format(pr_id))
                stop_all_jobs(launched_queues, queue_instance, running_builds)
                raise CIJobFail('New commit arrived: {0}'.format(commit_sha))

        # We update running builds and queues
        for queue_item in launched_queues:
            try:
                queue_item.poll()
                build = queue_item.get_build()
            except NotBuiltYet:
                logger.info('Item not built yet')
            except HTTPError:
                log.info('HTTP Error when querying idem {0}'.format(queue_item))
            else:
                running_builds.append(build)
                launched_queues.remove(queue_item)

        # Check if the builds are running + it updates their status
        running_builds, stopped_builds = split_on_condition(running_builds,
                                                            lambda build: build.is_running())

        if i % 20 == 0:
            log.info('Launched: {0}'.format(launched_queues))
            log.info('Running: {0}'.format(running_builds))
            log.info('Stopped: {0}'.format(stopped_builds))
        i += 1

        # We check for the second stopping condition, which is that a build failed
        failed_builds = filter(lambda build: not is_good(build), stopped_builds)
        if failed_builds:
            stop_all_jobs(launched_queues, queue_instance, running_builds)
            raise CIJobFail('Some builds failed: {0}'.format(failed_builds))

        # Sleep between 0.5 and 1s
        time.sleep(0.5 + random.randint(0, 500) / 1000)


def stop_all_jobs(launched_queues, queue_instance, running_builds):
    """Stop all jobs in queue or in the build pipe

    Args:
        launched_queues (list): list of queue items
        queue_instance (instance of jenkins.queues.Queue): instance of a Queue
        running_builds (list): list of running builds
    """
    # We delete queued jobs
    for queue_item in launched_queues:
        queue_instance.delete_item(queue_item)
    # And we stop running builds
    for build in running_builds:
        build.stop()


def launch_queue(jenkins_instance,
                 commit_id,
                 committer,
                 package,
                 git_branch,
                 type_build,
                 test_env_path):
    """Launch a CI queue

    Returns:
        jenkinsapi.queue.QueueItem: QueueItem instance
    """
    if type_build == 'artifactmaster':
        job_name = '{0} - {1}'.format(package, 'artifact')
    else:
        job_name = '{0} - {1}'.format(package, type_build)
    job = jenkins_instance[job_name]
    job_params = get_job_params(commit_id,
                                committer,
                                package,
                                git_branch,
                                type_build,
                                test_env_path)
    # Get a Queue instance which represents the new job build in the queue
    q = job.invoke(build_params=job_params)
    return q


def get_job_params(commit_id,
                   committer,
                   package,
                   git_branch,
                   type_build,
                   test_env_path):
    """Build the parameters to send to Jenkins jobs

    Args:
        commit_id:
        committer:
        package:
        git_branch:
        type_build:
        test_env_path:

    Returns:
    """
    composition_parameters = ('{0}:@:{1}'
                              .format(package,
                                      serialize(build_params_to_context(package, type_build)
                                                )
                                      )
                              )
    if type_build == 'test':
        job_params = {'package_name': composition_parameters,
                      'branch': ':@:'.join([git_branch,
                                            test_env_path,
                                            commit_id]),
                      'committer': committer}
    elif type_build == 'artifact':
        job_params = {'package_name': composition_parameters,
                      'branch': ':@:'.join([git_branch,
                                            test_env_path,
                                            commit_id]),
                      'committer': committer}
    elif type_build == 'artifactmaster':
        job_params = {'package_name': package,
                      'branch': git_branch,
                      'committer': committer}
    else:
        raise CIJobFail('Unknown type build {0}'.format(type_build))
    return job_params


def build_params_to_context(package, type_build):
    """Output a status message describing the build

    Args:
        package (str): Package being built
        type_build (str): type of build

    Returns:
        str: Message to use in Github status.
    """
    return ('{package} ({type_build})'
            .format(package=package, type_build=type_build))


def context_to_build_params(context):
    """Parse a Github status into the build parameters

    Args:
        context (basicstring): Github status message.

    Returns:
        package, type_build

    Raises:
        ValueError: Could not decode context
    """
    regex = re.compile('([\w\s-]+) \(([\w\s-]+)\)')
    groups = regex.findall(context)
    if not groups or len(groups[0]) != 2:
        raise ValueError('Could not decode context: "{0}". Regex groups: {1}'.format(context, groups))
    package, type_build = groups[0]
    return package, type_build



from __future__ import division

import random
import time

from loktar.log import Log


def retry(call):
    """Decorator used to retry a function

    Args:
        call: function to be retried

    Returns:
        function equipped with the decorator
    """

    def _retry(*args, **kwargs):
        """Fake doc"""
        logger = Log()
        last_exception = None
        multiplier = 1.5
        retry_interval = 0.5
        randomization_factor = 0.5
        total_sleep_time = 0
        max_sleep_time = 3 * 60

        request_nb = 0
        # Capped to 20 minutes
        while total_sleep_time < max_sleep_time:
            try:
                return call(*args, **kwargs)
            except Exception as exc:
                # Inspired from https://developers.google.com/api-client-library/java/google-http-java-client/reference/
                # 1.20.0/com/google/api/client/util/ExponentialBackOff
                next_retry_sleep = (multiplier ** request_nb *
                                    (retry_interval *
                                     (random.randint(0, int(2 * randomization_factor * 1000)) / 1000 +
                                      1 - randomization_factor)))

                total_sleep_time += next_retry_sleep
                request_nb += 1

                time.sleep(next_retry_sleep)

                last_exception = exc
                logger.info('Got an exception: {0}. Slept ({1} seconds / {2} seconds)'
                            .format(exc,
                                    total_sleep_time,
                                    max_sleep_time))
        logger.info('Max sleep time exceeded, raising exception.')
        raise last_exception

    # Keep the doc
    _retry.__doc__ = call.__doc__

    return _retry



from github import Github

from loktar.decorators import retry
from loktar.environment import GITHUB_INFO
from loktar.log import Log


@retry
def define_job_status_on_github_commit(commit_id, state, target_url, context, description=None, user=None, password=None, organization=None, repository=None):
    """Define the status of a job on a Github commit.

    Args:
        commit_id (str or int): SHA or commit id.
        state: One of 'pending', 'success', 'error', 'failure'.
        target_url (str): the URL that can be clicked in the Pull Request.
        context: The header of the status.
        description (Optional[str]): Optional description. Defaults according to ``state``.
        user (str, None): the user who post the status
        password (str, None): the user password
        organization (str, None): the organization of the repository target
        repository (str, None): the repository target
    """
    logger = Log()
    if commit_id is not None:
        connexion = Github(GITHUB_INFO['login']['user'] if GITHUB_INFO['login']['user'] is not None else user,
                           GITHUB_INFO['login']['password']
                           if GITHUB_INFO['login']['password'] is not None else
                           password)

        repository = connexion.get_organization(GITHUB_INFO['organization']
                                                if GITHUB_INFO['organization'] is not None else
                                                organization)\
                              .get_repo(GITHUB_INFO['repository']
                                        if GITHUB_INFO['repository'] is not None else
                                        repository)

        if description is None:
            if state == 'pending':
                description = GITHUB_INFO['notification']['pending']
            elif state == 'success':
                description = GITHUB_INFO['notification']['success']
            elif state == 'error':
                description = GITHUB_INFO['notification']['error']
            elif state == 'failure':
                description = GITHUB_INFO['notification']['failure']
            else:
                description = GITHUB_INFO['notification']['unknown']

        # Maximum length for description is 140 characters
        if len(description) > 140:
            description = description[:140 - 3] + '...'

        repository.get_commit(sha=commit_id).create_status(state=state,
                                                           target_url=target_url,
                                                           description=description,
                                                           context=context)

        logger.info('The job status is: {0}'.format(state))
    else:
        logger.info('Notifcation on github off')



import boto.s3 as s3

from loktar.decorators import retry
from loktar.environment import AWS
from loktar.exceptions import UnknownStorageMethod


@retry
def _store_test_env_on_s3(target, aws_region, aws_bucket):
    """Store an archive on aws s3

    Args:
        target (str): this is the archive target (eg: 1e827e44-f0a3-4f1f-a804-a28daff052af.tar.gz)
        aws_region (str, None): this is the aws region to connect for the bucket,
            aws_region can be set by a variable environment AWS_REGION, this variable it takes in first if it set
            can be set at None
        aws_bucket (str, None): this is the bucket name where the archive will be stored
            aws_bucket can be set by a variable environment AWS_BUCKET, this variable it takes in first if it set


    Returns
        A string who debribes the archive (eg: s3:@:foobar foobar here is the archive name)
    """
    key_name = target.split("/")[-1]
    connexion = s3.connect_to_region(AWS["REGION"] if AWS["REGION"] is not None else aws_region)
    bucket = connexion.get_bucket(AWS["BUCKET"] if AWS["BUCKET"] is not None else aws_bucket)
    key = bucket.new_key(key_name)
    key.set_contents_from_filename(target)

    return "s3:@:{0}".format(key_name)


def store_test_env(store_type, target, **kwargs):
    """Store the archive for launching test

    Args:
        store_type (str): define the storage type
        target (str): this is the archive target
        kwargs (dict): have some information around the storing method (eg: s3 need the bucket and region information)

    Returns
        A string who begin with the method for get back the archive and the archive name (eg: s3:@:foobar)
    """
    if store_type == "s3":
        return _store_test_env_on_s3(target, kwargs.get("aws_region", None), kwargs.get("aws_bucket", None))
    else:
        raise UnknownStorageMethod


@retry
def _get_back_test_env_from_s3(target, aws_region, aws_bucket):
    """Get back an archive from aws s3

        Args:
            target (str): this is the archive target
            aws_region (str, None): this is the aws region to connect for the bucket,
                aws_region can be set by a variable environment AWS_REGION, this variable it takes in first if it set
                can be set at None
            aws_bucket (str, None): this is the bucket name where the archive will be stored
                aws_bucket can be set by a variable environment AWS_BUCKET, this variable it takes in first if it set


        Returns
            A string who indicates where the archive is stored (eg: s3:@:foobar foobar here is the archive name)
    """
    test_env_location_ci = "/tmp/{0}".format(target)
    connexion = s3.connect_to_region(AWS["REGION"] if AWS["REGION"] is not None else aws_region)
    bucket = connexion.get_bucket(AWS["BUCKET"] if AWS["BUCKET"] is not None else aws_bucket)
    key = bucket.get_key(target)
    key.get_contents_to_filename(test_env_location_ci)

    return test_env_location_ci


def get_back_test_env(test_env_location, **kwargs):
    """Get back an archive from a source like aws s3, azure blob storage, remote dir ...

    Args:
        test_env_location (str): give information on the storage and the location of the archive

    Raises:
        UnknownStorageMethod: The storage method is not recognized
    """
    storage_type, archive = test_env_location.split(":@:")
    if storage_type == "s3":
        return _get_back_test_env_from_s3(archive, kwargs.get("aws_region", None), kwargs.get("aws_bucket", None))
    else:
        raise UnknownStorageMethod
